import json
import time
from dataclasses import dataclass

from loguru import logger

from ai_assistant_manager.chats.chat_response import MessageWithAnnotations

from ..clients.openai_api import OpenAIClient
from ..timer.timer import timer
from .chat_response import Annotation, ChatResponse

TOOL_CALL_PREFIX = "tc!"


class Chat:
    def __init__(self, client: OpenAIClient, assistant_id: str, *, thread_id: str | None = None):
        self.client = client
        self.assistant_id = assistant_id
        self.thread_id = thread_id

    def start(self):
        logger.info("Starting Chat")
        self.thread_id = self.thread_id or self.create_thread()
        logger.info(f"Thread ID: {self.thread_id}")

    def create_thread(self):
        return self.client.threads_create().id

    def send_user_message(self, message: str) -> ChatResponse:
        self.client.messages_create(
            self.thread_id,
            self.remove_tool_call_from_message(message),
            "user",
        )
        tokens = self.run_thread(self.should_force_tool_call(message))
        last_message = self.last_message_with_annotations()
        return ChatResponse(
            message=last_message.message, annotation_files=last_message.annotation_files, token_count=tokens
        )

    @timer("Submit Tool Outputs")
    def submit_tool_outputs(self, run_id: str, tool_call_id: str, response: str) -> int:
        self.client.submit_tool_outputs_to_run(run_id, tool_call_id, self.thread_id, response)
        tokens = self._wait_for_run_to_complete(run_id)
        return ChatResponse(message=self.last_message(), annotation_files=[], token_count=tokens)

    @timer("Run Thread")
    def run_thread(self, should_force_tool_call: bool = False) -> int:
        run = self.client.runs_create(self.assistant_id, self.thread_id, should_force_tool_call)
        return self._wait_for_run_to_complete(run.id)

    def _wait_for_run_to_complete(self, run_id: str, *, step: float = 0.25, timeout_in_seconds: int = 120) -> int:
        timeout = timeout_in_seconds / step

        while timeout > 0:
            run = self.client.runs_retrieve(run_id, self.thread_id)

            if run.status in ["completed"]:
                return run.usage.total_tokens
            if run.status in ["requires_action"] and run.required_action.type == "submit_tool_outputs":
                tool_call = run.required_action.submit_tool_outputs.tool_calls[0]
                tool_call_id = tool_call.id
                name = tool_call.function.name
                arguments = tool_call.function.arguments
                raise RequiresActionException(
                    f"Run requires action with status: {run.status}",
                    data=ActionData(
                        run_id=run_id, tool_call_id=tool_call_id, name=name, arguments=json.loads(arguments)
                    ),
                )
            if run.status in ["failed", "expired", "cancelled"]:
                raise RuntimeError(f"Run failed with status: {run.status}")

            timeout -= 1
            time.sleep(step)

        raise RuntimeError(f"Run timed out after {timeout_in_seconds} seconds")

    def last_message(self) -> str:
        message_content = self._get_messages()[0].content[0]
        if not hasattr(message_content, "text"):
            raise RuntimeError("No text content found in the messages")

        return message_content.text.value

    def last_message_with_annotations(self) -> MessageWithAnnotations:
        message_content = self._get_messages()[0].content[0]

        if not hasattr(message_content, "text"):
            raise RuntimeError("No text content found in the messages")

        annotations: list[Annotation] = [
            Annotation(file_id=annotation.file_citation.file_id, text=annotation.text)
            for annotation in message_content.text.annotations
        ]
        file_names = [self.client.files_get(annotation.file_id).filename for annotation in annotations]

        text_with_annotations = message_content.text.value
        for index, annotation in enumerate(annotations):
            text_with_annotations = text_with_annotations.replace(annotation.text, f"[*{index + 1}]")

        return MessageWithAnnotations(message=text_with_annotations, annotation_files=file_names)

    def _get_messages(self):
        return self.client.messages_list(self.thread_id).data

    def remove_tool_call_from_message(self, message: str) -> str:
        return message.replace(TOOL_CALL_PREFIX, "", 1) if self.should_force_tool_call(message) else message

    def should_force_tool_call(self, message: str) -> bool:
        return message.startswith(TOOL_CALL_PREFIX)


@dataclass
class ActionData:
    run_id: str
    tool_call_id: str
    name: str
    arguments: dict


class RequiresActionException(Exception):
    def __init__(self, message: str, *, data: ActionData):
        super().__init__(message)
        self.data = data



from .env_variables import ENV_VARIABLES, set_env_variables


def test_reset_env_variables(tmp_path):
    env_file = tmp_path / ".env"
    env_file.write_text(
        "OPENAI_MODEL=test_model\n"
        "ASSISTANT_DESCRIPTION=test_description\n"
        "ASSISTANT_NAME=test_name\n"
        "BIN_DIR=test_bin\n"
        "DATA_DIR=test_data\n"
        "DATA_FILE_PREFIX=test_prefix\n"
    )

    set_env_variables(str(env_file))

    assert ENV_VARIABLES.assistant_description == "test_description"
    assert ENV_VARIABLES.assistant_name == "test_name"
    assert ENV_VARIABLES.bin_dir == "test_bin"
    assert ENV_VARIABLES.data_dir == "test_data"
    assert ENV_VARIABLES.data_file_prefix == "test_prefix"
    assert ENV_VARIABLES.openai_model == "test_model"



from dataclasses import dataclass


@dataclass
class ChatResponse:
    message: str
    annotation_files: list[str]
    token_count: int


@dataclass
class MessageWithAnnotations:
    message: str
    annotation_files: list[str]


@dataclass
class Annotation:
    text: str
    file_id: str



import os

from loguru import logger

from ai_assistant_manager.chats.chat import Chat
from ai_assistant_manager.named_bytes import NamedBytesIO

from ..clients.openai_api import OpenAIClient
from ..env_variables import ENV_VARIABLES

RETRIEVAL_TOOLS = [
    {"type": "file_search"},
]


class AssistantService:
    """
    Service class to manage AI assistants and their associated vector stores and files.
    This class interacts with the OpenAIClient to perform operations such as creating,
    finding, and deleting assistants and their related resources.
    """

    def __init__(
        self,
        client: OpenAIClient,
        *,
        prompt: str | None = None,
        assistant_name: str | None = None,
        data_file_prefix: str | None = None,
        tools: list[dict] = RETRIEVAL_TOOLS,
    ):
        self.client = client
        self.prompt = prompt
        self.assistant_name = assistant_name if assistant_name else ENV_VARIABLES.assistant_name
        self.data_file_prefix = data_file_prefix if data_file_prefix else self.assistant_name
        self.tools = tools

    def get_assistant_id(self) -> str | None:
        assistant_id = self._find_existing_assistant(self.assistant_name)
        if assistant_id:
            return assistant_id
        else:
            return self._create_assistant()

    def get_assistant_by_key(self, assistant_key: str) -> str | None:
        return self._find_existing_assistant(assistant_key)

    def build_assistant(
        self, assistant_name: str, prompt: str, vector_store_ids: list[str] = [], tools: list[dict] = RETRIEVAL_TOOLS
    ):
        logger.info(f"Creating new assistant {assistant_name}")
        return self.client.assistants_create(
            assistant_name,
            prompt,
            vector_store_ids,
            tools=tools,
        ).id

    def start_chat(self, assistant_id: str, thread_id: str | None) -> Chat:
        chat = Chat(
            self.client,
            assistant_id,
            thread_id=thread_id,
        )
        chat.start()

        return chat

    def add_file_contents_to_files(self, file_contents: NamedBytesIO):
        return self.client.files_create(file_contents, "assistants").id

    def _find_existing_assistant(self, assistant_key: str):
        assistants = self.client.assistants_list()
        return next(
            (
                assistant.id
                for assistant in assistants
                if assistant.name == assistant_key or assistant.id == assistant_key
            ),
            None,
        )

    def _create_assistant(self):
        logger.info(f"Creating new assistant {self.assistant_name}")
        return self.client.assistants_create(
            self.assistant_name,
            self.prompt,
            self.get_vector_store_ids(),
            tools=self.tools,
        ).id

    def get_vector_store_ids(self):
        return self._find_existing_vector_stores() or self.create_vector_stores()

    def _find_existing_vector_stores(self):
        vector_stores = self.client.vector_stores_list()
        return [
            vector_store.id
            for vector_store in vector_stores
            if vector_store.name and vector_store.name.startswith(self.data_file_prefix)
        ]

    def create_vector_stores(self, *, vector_store_name: str = None, file_ids: list[str] = None):
        logger.info("Creating new vector stores")

        retrieval_file_ids = file_ids or self.get_retrieval_file_ids()
        vector_store_name = vector_store_name or f"{self.data_file_prefix} vector store"

        return [self._validate_vector_stores(self.client.vector_stores_create(vector_store_name, retrieval_file_ids))]

    def _validate_vector_stores(self, vector_store_id: str):
        try:
            vector_store_files = self.client.vector_stores_files(vector_store_id)
            failed_files = [file.id for file in vector_store_files if file.status == "failed"]

            if not failed_files:
                return vector_store_id

            failed_retrieval_files = [self.client.files_get(file) for file in failed_files if file]
            failed_retrieval_file_names = [self._get_file_name(file.filename) for file in failed_retrieval_files]
            failed_file_paths = [
                file_path
                for file_path in self._get_file_paths()
                if self._get_file_name(file_path) in failed_retrieval_file_names
            ]

            [self.client.vector_stores_file_delete(vector_store_id, file_id) for file_id in failed_files]

            recreated_files = self._create_files(failed_file_paths)
            self.client.vector_stores_update(vector_store_id, recreated_files)

            return self._validate_vector_stores(vector_store_id)
        except Exception as e:
            logger.error(f"Error validating vector store {vector_store_id}: {e}")
            return self._validate_vector_stores(vector_store_id)

    def _get_file_name(self, file_path: str) -> str:
        return os.path.basename(file_path)

    def get_retrieval_file_ids(self):
        return self._find_existing_retrieval_files() or self.create_retrieval_files()

    def _find_existing_retrieval_files(self):
        files = self.client.files_list()
        return [file.id for file in files if file.filename.startswith(self.data_file_prefix)]

    def create_retrieval_files(self):
        logger.info("Creating new retrieval files")
        file_paths = self._get_file_paths()
        return self._create_files(file_paths)

    def _get_file_paths(self):
        return [
            os.path.join(root, file)
            for (root, _, files) in os.walk("bin")
            for file in files
            if not file.endswith(".DS_Store")
        ]

    def _create_files(self, file_paths: list[str]):
        return [self._create_file(file_path) for file_path in file_paths]

    def _create_file(self, file_path: str):
        with open(file_path, "rb") as file:
            return self.client.files_create(file, "assistants").id

    def delete_assistant(self):
        logger.info(f"Removing existing {self.assistant_name} and retrieval files")

        if assistant_id := self._find_existing_assistant(self.assistant_name):
            self.client.assistants_delete(assistant_id)
        if vector_store_ids := self._find_existing_vector_stores():
            for vector_store_id in vector_store_ids:
                self.client.vector_stores_delete(vector_store_id)
        if file_ids := self._find_existing_retrieval_files():
            for file_id in file_ids:
                self.client.files_delete(file_id)



import json
import os
from dataclasses import asdict

from dateutil import parser
from loguru import logger

from ai_assistant_manager.encoding import UTF_8
from ai_assistant_manager.env_variables import ENV_VARIABLES

from ..content_data import ContentData
from ..exporter import create_dir, does_data_exist


class DirectoryExporter:
    def __init__(self, directory: str):
        self.directory = directory

    def export(self):
        if does_data_exist(self.get_file_path()):
            logger.info(f"Directory '{self.directory}' data exists. Skipping export.")
            return

        logger.info(f"Exporting directory '{self.directory}' data")
        create_dir(self.get_dir_path(), self.get_file_path())
        self.write_data()

    def write_data(self):
        data = self.load()
        data_as_dicts = {data.title: asdict(data) for data in data}
        json_data = json.dumps(data_as_dicts)

        with open(self.get_file_path(), "w", encoding=UTF_8) as file:
            file.write(json_data)

        logger.info(f"Directory '{self.directory}' data written to file: {self.get_file_path()}")

    def load(self):
        files = os.listdir(self.get_data_dir_path())
        return [self.file_load(filename) for filename in files]

    def file_load(self, filename: str) -> ContentData:
        file_id = filename[:3]
        name, _ = os.path.splitext(filename)
        title = name[3:].strip()

        with open(os.path.join(self.get_data_dir_path(), filename), "r", encoding=UTF_8) as file:
            lines = file.readlines()

        body = "\n".join([line.strip() for line in lines[1:]])
        date = parser.parse(lines[0].strip()).isoformat()

        return ContentData(id=file_id, title=title, body=body, date=date)

    def get_dir_path(self) -> str:
        return os.path.join(ENV_VARIABLES.bin_dir, self.directory)

    def get_file_path(self) -> str:
        return os.path.join(self.get_dir_path(), f"{ENV_VARIABLES.data_file_prefix} - {self.directory}.json")

    def get_data_dir_path(self) -> str:
        return os.path.join(ENV_VARIABLES.data_dir, self.directory)
